# This file implements a generic SQLite database data viewer/editor.
#

# Everything in a private namespace
namespace eval ::dbedit {

# proc:    ::dbedit::start DBOBJECT ARGS...
# title:   Launch a new db editor dialog box.
#
# Any number of editor dialog boxes can be active at one time.  Options
# are as follows:
#
#    -table TABLENAME       Start on the designated table
#    -where SQLEXPR         Start on the row indicated by SQLEXPR
#    -parent WINDOW         Center the dialog over WINDOW
#    -callback TCLPROC      Callbacked invoked after each change
#
proc start {dbname args} {
  set i 1
  while {[winfo exists .dbedit$i]} {incr i}
  set w .dbedit$i
  upvar #0 ::dbedit::v.dbedit$i v
  unset -nocomplain v
  set v(db) $dbname
  toplevel $w
  wm title $w "Database Editor"
  wm resizable $w 1 0
  array set ax $args
  if {[info exists ax(-table)]} {
    set v(curtab) $ax(-table)
  }
  if {[info exists ax(-callback)]} {
    set v(callback) $ax(-callback)
  }

  frame $w.bb
  pack $w.bb -side bottom -fill x -padx 5 -pady 5
  button $w.bb.close -text Close -command "::dbedit::_close $w"
  button $w.bb.del -text Delete -command "::dbedit::_delete $w"
  button $w.bb.save -text Save -command "::dbedit::_save $w"
  button $w.bb.revert -text Revert -command "::dbedit::_load_row $w"
  button $w.bb.new -text New -command "::dbedit::_new $w"
  button $w.bb.next -text Next
  button $w.bb.prev -text Prev
  pack $w.bb.close $w.bb.del $w.bb.save $w.bb.revert $w.bb.new \
       $w.bb.prev $w.bb.next -side right -expand 1 -padx 2
  if {[info proc ::sqlitecon::create]!=""} {
    button $w.bb.shell -text Shell -command \
        "::sqlitecon::create $w.con sqlite> {SQLite Console} $dbname"
    pack $w.bb.shell -side right -expand 1 -padx 2 -after $w.bb.close
  }
 

  frame $w.t
  pack $w.t -side top -fill x
  label $w.t.l1 -text Table:
  pack $w.t.l1 -side left
  tk_optionMenu  $w.t.m1 ::dbedit::v${w}(curtab) notused
  pack $w.t.m1 -side left
  label $w.t.l2 -text Pattern:
  pack $w.t.l2 -side left -padx {10 0}
  entry $w.t.e2 -bg white -fg black -relief sunken -width 20 \
     -textvariable ::dbedit::v${w}(pattern) 
  pack $w.t.e2 -side left -padx {0 5} -expand 1 -fill x

  frame $w.f
  pack $w.f -side top -fill both -expand 1 -padx 5 -pady 5
  _refill_table_menu $w
  trace variable ::dbedit::v${w}(curtab) w \
    "::dbedit::_rebuild_body_when_idle $w"
  trace variable ::dbedit::v${w} w \
    "::dbedit::_var_change $w"
  _rebuild_body $w
  if {[info exists ax(-where)] && [info exists ax(-table)]} {
    set qtab \"[string map {\" \"\"} $ax(-table)]\"
    set rowid [$dbname one "SELECT rowid FROM $qtab WHERE $ax(-where)"]
    if {[string is integer -strict $rowid]} {
      _goto_row $w $rowid
    }
  }

  if {[info exists ax(-parent)]} {
    set p $ax(-parent)
    wm withdraw $w
    ::update
    set ww [winfo reqwidth $w]
    set wh [winfo reqheight $w]
    set px [winfo rootx $p]
    set py [winfo rooty $p]
    set pw [winfo width $p]
    set ph [winfo height $p]
    set x [expr {$px+($pw-$ww)/2}]
    set y [expr {$py+($ph-$wh)/2}]
    if {$x<0} {set x 0}
    if {$y<0} {set y 0}
    wm geometry $w +$x+$y
    wm deiconify $w
  }
}

###########################################################################
# Public interface above.  Private data below
###########################################################################

# Called when the user presses the Close button on the main dialog window.
#
proc _close w {
  unset -nocomplain ::dbedit::v$w
  destroy $w
}

# Adjust the table selection menu so that it contains a list of all
# current tables in the database and is the right width to hold all of
# the table names.
#
proc _refill_table_menu w {
  upvar #0 ::dbedit::v$w v
  $w.t.m1.menu delete 0 end
  set maxlen 0
  set all {}
  set i 0
  set v(tablist) {}
  $v(db) eval {
     SELECT name FROM sqlite_master WHERE type='table'
     UNION
     SELECT name FROM sqlite_temp_master WHERE type='table'
     ORDER BY name COLLATE NOCASE
  } {
    lappend all $name
    $w.t.m1.menu add radiobutton -label $name \
        -variable ::dbedit::v${w}(curtab) -value $name \
        -columnbreak [expr {$i>0 && $i%20==0}]
    lappend v(tablist) $name
    set len [string length $name]
    if {$len>$maxlen} {set maxlen $len}
    incr i
  }
  $w.t.m1 config -width [expr {$maxlen+4}]
  if {![info exists v(curtab)] || [lsearch $all $v(curtab)]<0} {
    set v(curtab) [lindex $all 0]
  }
  _rebuild_body_when_idle $w
}

# Rebuild the part of the display that shows all of the columns in
# the table.  This happens whenever the table name changes.
#
proc _rebuild_body w {
  upvar #0 ::dbedit::v$w v
  if {![winfo exists $w]} return
  catch {after cancel $v(rebuild-pending)}
  unset -nocomplain v(rebuild-pending)
  set qtab '[string map {' ''} $v(curtab)]'
  bind Dbeditlabel <1> {::dbedit::_click %W %X %Y}
  foreach x [winfo children $w.f] {destroy $x}
  set pkey_type {}
  set pkey_name {}
  set pkey_cid {}
  $v(db) eval "PRAGMA foreign_key_list($qtab)" {
    set fkey_table($from) $table
    if {$to!=""} {
      set fkey_to($from) $to
    } else {
      set qto '[string map {' ''} $table]'
      $v(db) eval "PRAGMA table_info($qto)" {
         if {$pk} {
           set fkey_to($from) $name
           break
         }
      }
    }
  }
  $v(db) eval "PRAGMA table_info($qtab)" {
    set x $w.f.x$cid
    label ${x}_label -text $name
    bindtags ${x}_label [concat Dbeditlabel [bindtags ${x}_label]]
    grid ${x}_label -row $cid -column 3 -sticky en
    entry ${x}_entry -bg white -fg black -relief sunken -width 30 \
      -textvariable ::dbedit::v${w}($cid)
    set v($cid) {}
    grid ${x}_entry -row $cid -column 4 -sticky ewn
    if {$pk} {
      append pkey_type $type
      lappend pkey_name $name
      lappend pkey_cid $cid
    }
    grid rowconfig $w.f $cid -weight 0
    if {[info exists fkey_to($name)]} {
      button ${x}_ref -text Ref -padx 0 -pady 0 -bd 1 -command \
        [list ::dbedit::_ref_click $w $fkey_to($name) $fkey_table($name) $cid]
      grid ${x}_ref -row $cid -column 5 -sticky ew
    }
  }
  grid rowconfig $w.f $cid -weight 1
  if {[string tolower $pkey_type]=="integer"} {
    set v(intpkey) [lindex $pkey_cid 0]
  } else {
    set v(intpkey) {}
  }
  set v(ncol) [expr {$cid+1}]
  grid columnconfig $w.f 4 -weight 1
  menu $w.f.m -tearoff 0
  $w.f.m add command -label Search -command "::dbedit::_search $w"
  $w.f.m add command -label Edit -command "::dbedit::_edit $w"
  $w.f.m add command -label Index -command "::dbedit::_index $w"
  set minrow [$v(db) one "SELECT min(rowid) FROM $qtab"]
  if {$minrow==""} {set minrow 0}
  set v(rowid) $minrow
  _enable_next_prev $w
  _load_row $w
}

# Call _rebuild_body as an idle callback.
#
proc _rebuild_body_when_idle {w args} {
  upvar #0 ::dbedit::v$w v
  if {![info exists v(rebuild-pending)]} {
    set v(rebuild-pending) [after idle "::dbedit::_rebuild_body $w"]
  }
}

# This routine is called when the user clicks on one of "Ref" buttons
# of the database editor.  Bring up a new editor window on the entry
# that is referred to.
#
proc _ref_click {w field table cid} {
  upvar #0 ::dbedit::v$w v
  set val $v($cid) 
  set qval '[string map {' ''} $val]'
  set qfld \"[string map {\" \"\"} $field]\"
  start $v(db) -table $table -where $qfld=$qval
}

# This routine is called when the user clicks on one of the column
# labels in the main body of the window.  Show a popup menu of things
# to do with that column.
#
proc _click {W x y} {
  set w [winfo toplevel $W]
  upvar #0 ::dbedit::v$w v
  regexp {x([0-9]+)_label$} $W all cid
  set v(curcid) $cid
  tk_popup $w.f.m $x $y
}

# Based on the current setting of $v(rowid), enable or disable
# the Next and the Prev buttons.
#
proc _enable_next_prev w {
  upvar #0 ::dbedit::v$w v
  set rowid $v(rowid)
  set qtab '[string map {' ''} $v(curtab)]'
  set x [$v(db) one "SELECT rowid FROM $qtab WHERE rowid<\$rowid
                     ORDER BY rowid DESC"]
  if {$x==""} {
    $w.bb.prev config -state disabled
  } else {
    $w.bb.prev config -state normal -command "::dbedit::_goto_row $w $x"
  }
  set x [$v(db) one "SELECT rowid FROM $qtab WHERE rowid>\$rowid
                     ORDER BY rowid"]
  if {$x==""} {
    $w.bb.next config -state disabled
  } else {
    $w.bb.next config -state normal -command "::dbedit::_goto_row $w $x"
  }
}

# Load the current row into the body of the window.
#
proc _load_row w {
  upvar #0 ::dbedit::v$w v
  set rowid $v(rowid)
  set qtab '[string map {' ''} $v(curtab)]'
  set i 0
  foreach x [$v(db) eval "SELECT * FROM $qtab WHERE rowid=\$rowid"] {
    set v($i) $x
    incr i
  }
  if {$i==0} {
    while {[info exists v($i)]} {
      set v($i) {}
      incr i
    }
    $w.bb.del config -state disabled
    if {[winfo exists $w.f.lb]} {
      $w.f.lb selection clear 0 end
    }
  } else {
    if {[winfo exists $w.f.lb]} {
      $w.f.lb selection clear 0 end
      set i [lsearch [$w.f.lb get 0 end] $v($v(idxcid))]
      if {$i>=0} {
        $w.f.lb selection set $i
        $w.f.lb see $i
      }
    }
    $w.bb.del config -state normal
  }
  $w.bb.new config -state normal
  $w.bb.save config -state disabled
  $w.bb.revert config -state disabled
}

# Change to the rowid specified
#
proc _goto_row {w rowid} {
  upvar #0 ::dbedit::v$w v
  set v(rowid) $rowid
  _load_row $w
  _enable_next_prev $w
}

# Create an index of values for the indicated field.
#
proc _index {w} {
  upvar #0 ::dbedit::v$w v
  set field [$w.f.x$v(curcid)_label cget -text]
  set v(idxcid) $v(curcid)
  set v(idxfld) $field
  if {![winfo exists $w.f.lb]} {
    listbox $w.f.lb -yscrollcommand "$w.f.sb set" -exportsel 0
    grid $w.f.lb -row 0 -column 1 -rowspan $v(ncol) -sticky ns
    bind $w.f.lb <ButtonRelease-1> "::dbedit::_index_select $w"
    scrollbar $w.f.sb -orient vertical -command "$w.f.lb yview"
    grid $w.f.sb -row 0 -column 2 -rowspan $v(ncol) -sticky ns
  }
  _index_refill $w
}

# Refill the index listbox, if it exists.
#
proc _index_refill w {
  if {![winfo exists $w.f.lb]} return
  upvar #0 ::dbedit::v$w v
  $w.f.lb delete 0 end
  set qtab '[string map {' ''} $v(curtab)]'
  set qfield \"[string map {\" \"\"} $v(idxfld)]\"
  set sql "SELECT DISTINCT coalesce($qfield,'') AS x FROM $qtab ORDER BY 1"
  set curval $v($v(curcid))
  $v(db) eval $sql {
    $w.f.lb insert end $x
    if {$x==$curval} {
      $w.f.lb selection clear 0 end
      $w.f.lb see end
      $w.f.lb selection set end
    }
  }
  set n [$w.f.lb index end]
  if {$n>=2000} {
    $w.f.lb config -height 20
  } elseif {$n>=200} {
    $w.f.lb config -height 15
  }
}

# Called when the user clicks on the index listbox.
#
proc _index_select w {
  upvar #0 ::dbedit::v$w v
  if {![winfo exists $w.f.lb]} return
  set i [$w.f.lb cursel]
  if {$i==""} return
  set val [$w.f.lb get $i]
  set qtab '[string map {' ''} $v(curtab)]'
  set qfield \"[string map {\" \"\"} $v(idxfld)]\"
  set sql "SELECT rowid FROM $qtab WHERE CAST($qfield AS text)=\$val"
  set rowid [$v(db) one "$sql AND rowid>$v(rowid)"]
  if {$rowid==""} {
    set rowid [$v(db) one $sql]
  }
  if {$rowid!=""} {
    _goto_row $w $rowid
  }
}

# This routine is called when the user chooses the Edit option from the
# pop-up menu on the table column names.
#
proc _edit w {
  upvar #0 ::dbedit::v$w v
  set field [$w.f.x$v(curcid)_label cget -text]
  toplevel $w.ed
  wm protocol $w.ed WM_DELETE_WINDOW "$w.ed.can invoke"
  wm title $w.ed "Edit field \"$field\""
  frame $w.ed.b
  pack $w.ed.b -side bottom -fill x
  button $w.ed.b.can -text Cancel -width 6 \
      -command [list set ::dbedit::v${w}(continue) 0]
  pack $w.ed.b.can -side left -padx 5 -pady 5 -expand 1
  button $w.ed.b.cut -text Cut -width 6 \
      -command "::dbedit::_cut_text $w.ed.t"
  pack $w.ed.b.cut -side left -padx 5 -pady 5 -expand 1
  button $w.ed.b.copy -text Copy -width 6 \
      -command "::dbedit::_copy_text $w.ed.t"
  pack $w.ed.b.copy -side left -padx 5 -pady 5 -expand 1
  button $w.ed.b.paste -text Paste -width 6 \
      -command "::dbedit::_paste_text $w.ed.t"
  pack $w.ed.b.paste -side left -padx 5 -pady 5 -expand 1
  button $w.ed.b.ok -text OK -width 6 \
      -command [list set ::dbedit::v${w}(continue) 1]
  pack $w.ed.b.ok -side left -padx 5 -pady 5 -expand 1
  set ::$w {}
  text $w.ed.t -bg white -fg black -yscrollcommand [list $w.ed.sb set] \
        -wrap char -exportselection 0
  pack $w.ed.t -side left -fill both -expand 1
  scrollbar $w.ed.sb -orient vertical -command [list $w.ed.t yview]
  pack $w.ed.sb -side left -fill y
  $w.ed.t insert end $v($v(curcid))
  ::tk::PlaceWindow $w.ed widget $w

  set v(continue) {}
  vwait ::dbedit::v${w}(continue)

  if {$v(continue)} {
    set v($v(curcid)) [string trimright [$w.ed.t get 1.0 end]]
  }
  destroy $w.ed
}

# Copy, Cut, or Paste the selection from text widget $t into the clipboard.
#
proc _copy_text {t} {
  if {![catch {$t get sel.first sel.last} text]} {
     clipboard clear -displayof $t
     clipboard append -displayof $t $text
  }
}
proc _cut_text {t} {
  ::dbedit::_copy_text $t
  catch {$t delete sel.first sel.last}
}
proc _paste_text {t} {
  if {[catch {selection get -displayof $t -selection CLIPBOARD} topaste]} {
    return
  }
  $t insert insert $topaste
}


# This routine is called whenever any entry in the state variable for
# a dialog box changes.
#
proc _var_change {w notused1 rownum notused2} {
  if {![string is integer $rownum]} return
  if {![winfo exists $w.bb.save]} return
  if {[$w.bb.save cget -state]=="normal"} return
  $w.bb.save config -state normal
  $w.bb.revert config -state normal
  $w.bb.new config -state disabled
  $w.bb.next config -state disabled
  $w.bb.prev config -state disabled
}

# Invoke the change callback routine, if there is one.
#
proc _callback {w rowid} {
  upvar #0 ::dbedit::v$w v
  if {[info exists v(callback)]} {
    uplevel #0 $v(callback) [list $v(curtab)] $rowid
  }
}

# This routine is called when the user presses the Delete button.
#
proc _delete w {
  set q [tk_messageBox -icon question -type yesno -parent $w -message \
              "Really delete this row?"]
  if {$q!="yes"} return
  upvar #0 ::dbedit::v$w v
  set oldrowid $v(rowid)
  if {[$w.bb.next cget -state]=="normal"} {
    $w.bb.next invoke
  } elseif {[$w.bb.prev cget -state]=="normal"} {
    $w.bb.prev invoke
  } else {
    set v(rowid) {}
    _load_row $w
  }
  set qtab '[string map {' ''} $v(curtab)]'
  $v(db) eval "DELETE FROM $qtab WHERE rowid=\$oldrowid"
  _callback $w $oldrowid
  _index_refill $w
  _enable_next_prev $w
}

# This routine is called when the user presses the New button
#
proc _new w {
  _goto_row $w {}
}

# This routine is called when the user presses the Save button
#
proc _save w {
  upvar #0 ::dbedit::v$w v
  if {$v(rowid)==""} {
    # Insert a new row
    set sql "INSERT INTO '[string map {' ''} $v(curtab)]' VALUES"
    set sep (
    for {set i 0} {$i<$v(ncol)} {incr i} {
      set val $v($i)
      if {$i==$v(intpkey) && $val==""} {
        append sql ${sep}NULL
      } elseif {[string is double -strict $val]} {
        append sql $sep$val
      } else {
        append sql $sep'[string map {' ''} $val]'
      }
      set sep ,
    }
    append sql )
  } else {
    # Update an existing row
    set sql "UPDATE '[string map {' ''} $v(curtab)]' "
    set sep {SET}
    set rowid $v(rowid)
    set newrowid $rowid
    for {set i 0} {$i<$v(ncol)} {incr i} {
      set val $v($i)
      set name [$w.f.x${i}_label cget -text]
      append sql "$sep $name="
      if {[string is double -strict $val]} {
        append sql $val
      } else {
        append sql '[string map {' ''} $val]'
      }
      if {$i==$v(intpkey)} {
        set newrowid $val
      }
      set sep ,
    }
    append sql " WHERE rowid=\$rowid"
  }
  if {[catch {$v(db) eval $sql} errmsg]} {
    tk_messageBox -icon error -type ok -parent $w -message $errmsg
  } else {
    if {![info exists newrowid]} {
      set newrowid [$v(db) last_insert_rowid]
    }
    _callback $w $newrowid
    _index_refill $w
    _goto_row $w $newrowid
  }
}

# Search for the next occurrance of Pattern.
#
proc _search w {
  set pattern [string trim [$w.t.e2 get]]
  if {$pattern==""} {
    tk_messageBox -icon error -type ok -parent $w -message \
        "Nothing to search for"
    return
  }
  set qpattern '%[string map {' ''} $pattern]%'
  upvar #0 ::dbedit::v$w v
  set field [$w.f.x$v(curcid)_label cget -text]
  set qfield \"[string map {\" \"\"} $field]\"
  set qtab '[string map {' ''} $v(curtab)]'
  set rowid $v(rowid)
  set sql "SELECT rowid FROM $qtab WHERE $qfield LIKE $qpattern"
  set newrowid [$v(db) one "$sql AND rowid>\$rowid ORDER BY rowid"]
  if {$newrowid==""} {
    set newrowid [$v(db) one "$sql AND rowid<=\$rowid ORDER BY rowid"]
  }
  if {$newrowid==""} {
    tk_messageBox -icon error -type ok -parent $w -message "Not found"
    return
  }
  _goto_row $w $newrowid
}

# End of the ::dbedit namespace
}
